from typing import Callable, Union, Type, Any

from .event_metadata import EventMetadata, EventData
from nestipy.metadata import SetMetadata


def OnEvent(name: str) -> Callable[[Union[Type, Callable[..., Any]]], Any]:
    return SetMetadata(EventMetadata.Event, EventData(name=name, once=False))


def OnceEvent(name: str) -> Callable[[Union[Type, Callable[..., Any]]], Any]:
    return SetMetadata(EventMetadata.Event, EventData(name=name, once=True))



import dataclasses
from abc import abstractmethod, ABC
from typing import Any, TYPE_CHECKING

import dataclasses_json

from nestipy.common.exception.http import HttpException
from nestipy.common.exception.interface import ExceptionFilter

if TYPE_CHECKING:
    from nestipy.core.context.argument_host import ArgumentHost


class RPCErrorCode:
    OK = 0
    CANCELLED = 1
    UNKNOWN = 2
    INVALID_ARGUMENT = 3
    DEADLINE_EXCEEDED = 4
    NOT_FOUND = 5
    ALREADY_EXISTS = 6
    PERMISSION_DENIED = 7
    RESOURCE_EXHAUSTED = 8
    FAILED_PRECONDITION = 9
    ABORTED = 10
    OUT_OF_RANGE = 11
    UNIMPLEMENTED = 12
    INTERNAL = 13
    UNAVAILABLE = 14
    DATA_LOSS = 15
    UNAUTHENTICATED = 16


class RPCErrorMessage:
    OK = "The operation completed successfully."
    CANCELLED = "The operation was cancelled, typically by the caller."
    UNKNOWN = "An unknown error occurred."
    INVALID_ARGUMENT = "The client specified an invalid argument."
    DEADLINE_EXCEEDED = "The deadline expired before the operation could complete."
    NOT_FOUND = "Some requested entity (e.g., file or directory) was not found."
    ALREADY_EXISTS = "The entity that a client attempted to create already exists."
    PERMISSION_DENIED = "The caller does not have permission to execute the specified operation."
    RESOURCE_EXHAUSTED = ("Some resource has been exhausted, perhaps a per-user quota, or perhaps the entire file "
                          "system is out of space.")
    FAILED_PRECONDITION = ("The operation was rejected because the system is not in a state required for the "
                           "operation's execution.")
    ABORTED = ("The operation was aborted, typically due to a concurrency issue such as a sequencer check failure or "
               "transaction abort.")
    OUT_OF_RANGE = "The operation was attempted past the valid range."
    UNIMPLEMENTED = "The operation is not implemented or is not supported/enabled in this service."
    INTERNAL = "Internal errors. Means some invariants expected by the underlying system have been broken."
    UNAVAILABLE = "The service is currently unavailable."
    DATA_LOSS = "Unrecoverable data loss or corruption."
    UNAUTHENTICATED = "The request does not have valid authentication credentials for the operation."


@dataclasses_json.dataclass_json
@dataclasses.dataclass
class RpcException(HttpException):
    status_code: int
    message: str

    def __post_init__(self):
        super().__init__(self.status_code, self.message)

    def to_json(self):
        return {"status_code": self.status_code, "message": self.message}

    @classmethod
    def from_json(cls, data: dict):
        return cls(data.get("status_code"), data.get("message"))


class RpcExceptionFilter(ExceptionFilter, ABC):

    @abstractmethod
    async def catch(self, exception: "RpcException", host: "ArgumentHost") -> Any:
        pass



from typing import Type

from nestipy.common.decorator import Injectable

from nestipy.types_ import NextFn
from nestipy.common.middleware.interface import NestipyMiddleware
from nestipy.common.http_ import Request, Response


def helmet() -> Type:
    """
    Returns:
        HelmetMiddleware(Callable): Helmet middlewares for Nestipy
    """
    @Injectable()
    class HelmetMiddleware(NestipyMiddleware):
        async def use(self, req: Request, res: Response, next_fn: NextFn):
            result = await next_fn()
            res.header('X-Content-Type-Options', 'nosniff')
            res.header('X-Frame-Options', 'DENY')
            res.header('X-XSS-Protection', '1; mode=block')
            res.header('Strict-Transport-Security', 'max-age=31536000; includeSubDomains')
            res.header('Content-Security-Policy', "default-src 'self'")
            res.header('Referrer-Policy', 'no-referrer')
            return result

    return HelmetMiddleware



import uuid
from dataclasses import dataclass, field
from typing import TypeVar, Generic, Optional, Any, Callable, Union, Type, Awaitable

from nestipy.ioc import ModuleProviderDict
from nestipy.metadata import ModuleMetadata, Reflect

T = TypeVar('T')


@dataclass
class DynamicModule:
    module: Any
    exports: list = field(default_factory=lambda: [])
    imports: list = field(default_factory=lambda: [])
    providers: list = field(default_factory=lambda: [])
    controllers: list = field(default_factory=lambda: [])
    is_global: bool = False


class ConfigurableModuleBuilder(Generic[T]):
    def __init__(self):
        self._method_name = 'register'
        self._extras: Union[dict[str, Any], None] = None
        self._extras_process_callback: Union[Callable[[DynamicModule, dict[str, Any]], None], None] = None

    def set_extras(
            self,
            extras: dict[str, Any],
            extras_callback: Callable[[DynamicModule, dict[str, Any]], None]
    ):
        self._extras = extras
        self._extras_process_callback = extras_callback
        return self

    def set_method(self, name: str):
        self._method_name = name
        return self

    def _extra_return(self, dynamic_module: DynamicModule) -> DynamicModule:
        if self._extras is not None and self._extras_process_callback is not None:
            self._extras_process_callback(dynamic_module, self._extras)
        return dynamic_module

    def _create_dynamic_module(self, obj: Any, imports: list, provider: list) -> DynamicModule:
        dynamic_module = DynamicModule(
            obj,
            providers=provider + Reflect.get_metadata(obj, ModuleMetadata.Providers, []),
            exports=Reflect.get_metadata(obj, ModuleMetadata.Exports, []),
            imports=imports + Reflect.get_metadata(obj, ModuleMetadata.Imports, []),
            controllers=Reflect.get_metadata(obj, ModuleMetadata.Controllers, []),
            is_global=Reflect.get_metadata(obj, ModuleMetadata.Global, False)
        )
        return self._extra_return(dynamic_module)

    def build(self) -> tuple[Type, str]:
        MODULE_OPTION_TOKEN = f"{uuid.uuid4().hex}_TOKEN"

        def register(cls_: Any, options: Optional[T], extras: dict = None) -> DynamicModule:
            if extras is not None:
                self._extras = extras
            provider = ModuleProviderDict(
                token=MODULE_OPTION_TOKEN,
                value=options
            )
            return self._create_dynamic_module(cls_, [], [provider])

        def register_async(
                cls_: Any,
                value: Optional[T] = None,
                factory: Callable[..., Union[Awaitable, Any]] = None,
                existing: Union[Type, str] = None,
                use_class: Type = None,
                inject: list = None,
                imports: list = None,
                extras: dict = None
        ) -> DynamicModule:
            if extras is not None:
                self._extras = extras
            provider = ModuleProviderDict(
                token=MODULE_OPTION_TOKEN,
                factory=factory,
                inject=inject or [],
                use_class=use_class,
                existing=existing,
                value=value
            )
            return self._create_dynamic_module(cls_, imports or [], [provider])

        cls = type('ConfigurableModuleClass', (object,), {
            self._method_name: classmethod(register),
            f"{self._method_name}_async": classmethod(register_async)
        })
        return cls, MODULE_OPTION_TOKEN



from abc import ABC
from dataclasses import asdict
from typing import Callable, Any

from nestipy.microservice.client import ClientProxy
from nestipy.microservice.client.option import Transport
from nestipy.microservice.context import RpcRequest, RpcResponse, MICROSERVICE_CHANNEL
from nestipy.microservice.exception import RpcException, RPCErrorCode, RPCErrorMessage


class MicroServiceServer(ABC):
    def __init__(self, pubsub: ClientProxy):
        self._pubsub = pubsub
        self._subscriptions: list[tuple[str, Callable]] = []
        self._request_subscriptions: list[tuple[str, Callable]] = []

    async def listen(self):
        await self._pubsub.connect()
        await self._pubsub.subscribe(MICROSERVICE_CHANNEL)
        async for data in self._pubsub.listen():
            json_rep = await self._pubsub.option.serializer.deserialize(data)
            request = RpcRequest(**json_rep)
            if request.is_event():
                await self.handle_event(request)
            else:
                await self.handle_request(request=request)

    def get_transport(self) -> Transport:
        return self._pubsub.option.transport

    async def close(self):
        await self._pubsub.close()

    def subscribe(self, topic: str, callback: Callable):
        self._subscriptions.append((topic, callback))

    def unsubscribe(self, topic: str):
        for sub in self._subscriptions:
            if sub[0] == topic:
                self._subscriptions.remove(sub)

    def request_subscribe(self, topic: str, callback: Callable):
        self._request_subscriptions.append((topic, callback))

    def request_unsubscribe(self, topic: str):
        for sub in self._request_subscriptions:
            if sub[0] == topic:
                self._request_subscriptions.remove(sub)

    async def handle_request(self, request: RpcRequest) -> Any:
        # process pattern
        slave = await self._pubsub.slave()
        await slave.connect()
        for (pattern, callback) in self._request_subscriptions:
            if pattern == request.pattern:
                # resolve dependencies
                response = await callback(self, request)
                rpc_response = RpcResponse(
                    pattern=request.response_topic,
                    data=response,
                    status="success"
                )
                await slave.send_response(
                    request.response_topic,
                    await self._pubsub.option.serializer.serialize(asdict(rpc_response))
                )
                return
        rpc_response = await self._pubsub.option.serializer.serialize(
            asdict(
                RpcResponse(
                    pattern=request.pattern,
                    data=None,
                    status="error",
                    exception=RpcException(
                        status_code=RPCErrorCode.NOT_FOUND,
                        message=RPCErrorMessage.NOT_FOUND
                    )
                )
            )
        )
        await slave.send_response(
            request.response_topic,
            rpc_response
        )
        await slave.close()

    async def handle_event(self, request: RpcRequest):
        # process pattern
        for (pattern, callback) in self._subscriptions:
            if pattern == request.pattern:
                await callback(self, request)



from typing import Annotated

from nestipy.common import Module
from nestipy.core import DiscoverService
from nestipy.ioc import Inject


@Module()
class MicroserviceServerModule:
    discover: Annotated[DiscoverService, Inject()]



import inspect
from typing import Callable, Type, Any

from strawberry import type as strawberry_type, field as strawberry_field, mutation, subscription, Schema

from ..graphql_adapter import GraphqlAdapter
from ..graphql_asgi import GraphqlAsgi
from ..graphql_module import GraphqlOption
from ..strawberry.strawberry_asgi import StrawberryAsgi


class StrawberryAdapter(GraphqlAdapter):
    def raise_exception(self, e: Exception):
        raise e

    _schema: Schema = None

    def create_query_field_resolver(self, resolver: Callable) -> object:
        return strawberry_field(resolver)

    def create_mutation_field_resolver(self, resolver: Callable) -> object:
        return mutation(resolver)

    def create_subscription_field_resolver(self, resolver: Callable) -> object:
        return subscription(resolver)

    def mutate_handler(self, original_handler: Any, mutated_handler: Callable) -> Type:
        signature = inspect.signature(original_handler)
        return_annotation = signature.return_annotation
        # Set annotations for the new function
        if return_annotation is None:
            raise Exception(f"Method {original_handler.__name__} has nor return annotation")

        mutated_handler.__annotations__ = original_handler.__annotations__
        mutated_handler.__signature__ = signature
        return return_annotation

    def create_schema(self, *args, **kwargs):
        query = self.create_query()
        mutation_ = self.create_mutation()
        subscription_ = self.create_subscription()
        if query is None:
            def health() -> bool:
                return True

            query = type('Query', (), {'health': strawberry_field(health)})
        return Schema(
            query=strawberry_type(query),
            mutation=strawberry_type(mutation_) if mutation_ is not None else None,
            subscription=strawberry_type(subscription_) if subscription_ is not None else None,
            *args, **kwargs
        )

    def create_graphql_asgi_app(self, schema: Any, option: GraphqlOption) -> GraphqlAsgi:
        app_asgi = StrawberryAsgi(schema=schema, option=option)
        return app_asgi



import enum
from typing import Type, Callable, Union

from nestipy.dynamic_module import DynamicModule
from nestipy.ioc import NestipyContainer, ModuleProviderDict
from nestipy.metadata import ModuleMetadata, Reflect, RouteKey


class Scope(enum.Enum):
    Request = 'Request'
    Transient = 'Transient'
    Singleton = 'Singleton'


class Injectable:
    scope: Scope = None

    def __init__(self, scope: Scope = Scope.Singleton):
        self.scope = scope
        self.container = NestipyContainer.get_instance()

    def __call__(self, cls: Type) -> Type:
        match self.scope:
            case Scope.Transient:
                self.container.add_transient(cls)
            case Scope.Request:
                self.container.add_transient(cls)
            case _:
                self.container.add_singleton(cls)
        return cls


class Controller:
    def __init__(self, path: str = '/', **kwargs):
        self.path = path
        self.kwargs = kwargs
        self.container = NestipyContainer.get_instance()

    def __call__(self, cls, **kwargs):
        self.container.add_singleton(cls)
        # put path and kwargs in controller property
        Reflect.set_metadata(cls, RouteKey.path, self.path)
        Reflect.set_metadata(cls, RouteKey.kwargs, self.kwargs)
        return cls


class Module:
    providers: list[Union[Type, ModuleProviderDict]] = []
    controllers: list[Type] = []
    imports: list[Union[Type, Callable, ModuleProviderDict, DynamicModule]] = []
    exports: list[Union[Type, Callable, str]] = []
    is_global: bool = False

    def __init__(
            self,
            providers: list[Callable | ModuleProviderDict] = None,
            controllers: list[Callable] = None,
            imports: list[Union[Type, Callable, ModuleProviderDict, DynamicModule]] = None,
            exports: list[Union[Type, Callable, str]] = None,
            is_global: bool = False
    ):
        self.providers = providers or []
        self.controllers = controllers or []
        self.imports = imports or []
        self.exports = exports or []
        self.is_global = is_global
        self.container = NestipyContainer.get_instance()

    def __call__(self, cls: Type):
        Reflect.set_metadata(cls, ModuleMetadata.Providers, self.providers + getattr(cls, ModuleMetadata.Providers, []))
        Reflect.set_metadata(
            cls,
            ModuleMetadata.Controllers,
            self.controllers + getattr(cls, ModuleMetadata.Controllers, [])
        )
        Reflect.set_metadata(cls, ModuleMetadata.Imports, self.imports + getattr(cls, ModuleMetadata.Imports, []))
        Reflect.set_metadata(cls, ModuleMetadata.Exports, self.exports + getattr(cls, ModuleMetadata.Exports, []))
        Reflect.set_metadata(cls, ModuleMetadata.Global, self.is_global or getattr(cls, ModuleMetadata.Global, False))
        Reflect.set_metadata(cls, ModuleMetadata.Module, True)
        self.container.add_singleton(cls)
        return cls



from .client import ClientsModule, ClientsConfig, ClientModuleFactory, MicroserviceOption
from .client import RedisClientProxy, MqttClientProxy, ClientProxy, RabbitMQClientProxy, NatsClientProxy
from .client.option import MicroserviceClientOption, Transport
from .context import RpcResponse, RpcRequest
from .decorator import MessagePattern, EventPattern
from .dependency import Payload, Ctx, Client
from .exception import RpcException, RpcExceptionFilter, RPCErrorMessage, RPCErrorCode
from .serilaizer import JSONSerializer
from .server import MicroServiceServer

__all__ = [
    "ClientsModule",
    "ClientsConfig",
    "ClientProxy",
    "RedisClientProxy",
    "MqttClientProxy",
    "RabbitMQClientProxy",
    "NatsClientProxy",
    "MicroserviceOption",
    "MicroserviceClientOption",
    "Transport",
    "MessagePattern",
    "EventPattern",
    "RpcResponse",
    "RpcRequest",
    "RpcException",
    "RpcExceptionFilter",
    "MicroServiceServer",
    "Payload",
    "Ctx",
    "Client",
    "RPCErrorMessage",
    "RPCErrorCode",
    "ClientModuleFactory"
]



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from nestipy.common.http_ import Request, Response
    from nestipy.types_ import NextFn


class NestipyMiddleware(ABC):
    @abstractmethod
    async def use(self, req: "Request", res: "Response", next_fn: "NextFn"):
        """
        Args:
            req (Request): The http request
            res (Response): the http response
            next_fn (NextFn): th next function handler
        """
        pass



import os
from typing import Any

from minijinja import Environment

from nestipy.common.template import TemplateEngine


class MinimalJinjaTemplateEngine(TemplateEngine):
    env: Environment

    def __init__(self, template_dir: str):
        super().__init__(template_dir)
        self.env = Environment(
            loader=self.loader
        )

    def loader(self, name):
        segments = []
        for segment in name.split("/"):
            if "\\" in segment or segment in (".", ".."):
                return None
            segments.append(segment)
        try:
            path = os.path.join(self.template_dir, *segments)
            with open(path) as f:
                content = f.read()
                return content
        except (IOError, OSError):
            pass

    def render(self, template: str, context: dict) -> str:
        return self.env.render_template(template, **context)

    def render_str(self, string: str, context: dict) -> str:
        return self.env.render_str(string, **context)

    def get_env(self) -> Any:
        return self.env



"""
This module defines classes that can be used to generate OpenAPI Documentation
version 2.
https://swagger.io/specification/v2/
"""
from abc import ABC
from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, List, Optional, Type, Union

from .common import OpenAPIRoot

from .common import OpenAPIElement


def get_ref(ref_type: Union[str, Type]) -> str:
    if isinstance(ref_type, str):
        return f"#/definitions/{ref_type}"
    return f"#/definitions/{ref_type.__name__}"


class ParameterLocation(Enum):
    QUERY = "query"
    HEADER = "header"
    PATH = "path"
    FORM = "formData"
    BODY = "body"


class APIKeyLocation(Enum):
    QUERY = "query"
    HEADER = "header"


class ValueType(Enum):
    ARRAY = "array"
    BOOLEAN = "boolean"
    FILE = "file"
    INTEGER = "integer"
    NUMBER = "number"
    OBJECT = "object"
    STRING = "string"


class HeaderType(Enum):
    ARRAY = "array"
    BOOLEAN = "boolean"
    INTEGER = "integer"
    NUMBER = "number"
    STRING = "string"


class ValueItemType(Enum):
    ARRAY = "array"
    BOOLEAN = "boolean"
    INTEGER = "integer"
    NUMBER = "number"
    STRING = "string"


class ValueFormat(Enum):
    BINARY = "binary"
    BYTE = "byte"
    DATE = "date"
    DATETIME = "date-time"
    DOUBLE = "double"
    FLOAT = "float"
    INT32 = "int32"
    INT64 = "int64"
    PASSWORD = "password"


class CollectionFormat(Enum):
    CSV = "csv"
    SSV = "ssv"
    TSV = "tsv"
    PIPES = "pipes"
    MULTI = "multi"


class SecuritySchemeType(Enum):
    BASIC = "basic"
    APIKEY = "apiKey"
    OAUTH = "oauth2"
    OAUTH2 = "oauth2"


class OAuthFlowType(Enum):
    IMPLICIT = "implicit"
    PASSWORD = "password"
    APPLICATION = "application"
    ACCESS_CODE = "accessCode"


@dataclass
class Contact(OpenAPIElement):
    name: Optional[str] = None
    url: Optional[str] = None
    email: Optional[str] = None


@dataclass
class ExternalDocs(OpenAPIElement):
    url: str
    description: Optional[str] = None


@dataclass
class License(OpenAPIElement):
    name: str
    url: Optional[str] = None


@dataclass
class Info(OpenAPIElement):
    title: str
    version: str
    description: Optional[str] = None
    terms_of_service: Optional[str] = None
    contact: Optional[Contact] = None
    license: Optional[License] = None


@dataclass
class XML(OpenAPIElement):
    name: Optional[str] = None
    namespace: Optional[str] = None
    prefix: Optional[str] = None
    attribute: Optional[bool] = None
    wrapped: Optional[bool] = None


@dataclass
class Discriminator(OpenAPIElement):
    property_name: str
    mapping: Optional[Dict[str, str]] = None


@dataclass
class Schema(OpenAPIElement):
    type: Union[None, str, ValueType] = None
    format: Union[None, str, ValueFormat] = None
    required: Optional[List[str]] = None
    properties: Optional[Dict[str, "Schema"]] = None
    default: Optional[Any] = None
    deprecated: Optional[bool] = None
    example: Any = None
    external_docs: Optional[ExternalDocs] = None
    ref: Optional[str] = None
    title: Optional[str] = None
    description: Optional[str] = None
    max_length: Optional[float] = None
    min_length: Optional[float] = None
    maximum: Optional[float] = None
    minimum: Optional[float] = None
    nullable: Optional[bool] = None
    xml: Optional[XML] = None
    items: Optional["Schema"] = None
    enum: Optional[List[str]] = None
    discriminator: Optional[Discriminator] = None
    all_of: Optional[List[Union["Schema", "Reference"]]] = None
    any_of: Optional[List[Union["Schema", "Reference"]]] = None
    one_of: Optional[List[Union["Schema", "Reference"]]] = None
    not_: Optional[List[Union["Schema", "Reference"]]] = None


@dataclass
class Header(OpenAPIElement):
    type: HeaderType
    description: Optional[str] = None
    format: Optional[str] = None
    items: Optional["Items"] = None
    collection_format: Optional[CollectionFormat] = None
    default: Any = None
    maximum: Optional[float] = None
    minimum: Optional[float] = None
    exclusive_maximum: Optional[bool] = None
    exclusive_minimum: Optional[bool] = None
    enum: Optional[List[str]] = None
    max_length: Optional[int] = None
    min_length: Optional[int] = None
    pattern: Optional[str] = None
    max_items: Optional[int] = None
    min_items: Optional[int] = None
    unique_items: Optional[bool] = None
    multiple_of: Optional[float] = None


@dataclass
class Example(OpenAPIElement):
    summary: Optional[str] = None
    description: Optional[str] = None
    value: Any = None
    external_value: Optional[str] = None


@dataclass
class Reference(OpenAPIElement):
    ref: str

    def to_obj(self) -> Dict[str, str]:
        return {"$ref": self.ref}


@dataclass
class Encoding(OpenAPIElement):
    content_type: Optional[str] = None
    headers: Optional[Dict[str, Union[Header, Reference]]] = None
    style: Optional[str] = None
    explode: Optional[bool] = None
    allow_reserved: Optional[bool] = None


@dataclass
class Response(OpenAPIElement):
    description: str
    headers: Optional[Dict[str, Union[Header, Reference]]] = None
    schema: Optional[Schema] = None
    examples: Optional[Dict[str, Any]] = None


@dataclass
class Items(OpenAPIElement):
    type: ValueItemType
    format: Optional[ValueFormat] = None
    items: Optional["Items"] = None
    collection_format: Optional[CollectionFormat] = None
    default: Optional[str] = None
    maximum: Optional[float] = None
    minimum: Optional[float] = None
    exclusive_maximum: Optional[bool] = None
    exclusive_minimum: Optional[bool] = None
    enum: Optional[List[str]] = None
    max_length: Optional[int] = None
    min_length: Optional[int] = None
    pattern: Optional[str] = None
    max_items: Optional[int] = None
    min_items: Optional[int] = None
    unique_items: Optional[bool] = None
    multiple_of: Optional[float] = None


@dataclass
class Parameter(OpenAPIElement):
    name: str
    in_: ParameterLocation
    type: Optional[ValueType] = None
    format: Optional[ValueFormat] = None
    items: Optional[Items] = None
    collection_format: Optional[CollectionFormat] = None
    schema: Optional[Schema] = None
    description: Optional[str] = None
    allow_empty_value: Optional[bool] = None
    example: Optional[Any] = None
    examples: Optional[Dict[str, Union[Example, Reference]]] = None
    default: Optional[str] = None
    maximum: Optional[float] = None
    minimum: Optional[float] = None
    exclusive_maximum: Optional[bool] = None
    exclusive_minimum: Optional[bool] = None
    enum: Optional[List[str]] = None
    max_length: Optional[int] = None
    min_length: Optional[int] = None
    pattern: Optional[str] = None
    max_items: Optional[int] = None
    min_items: Optional[int] = None
    unique_items: Optional[bool] = None
    multiple_of: Optional[float] = None
    required: Optional[bool] = None


@dataclass
class SecurityRequirement(OpenAPIElement):
    name: str
    value: List[str]

    def to_obj(self):
        return {self.name: self.value}


@dataclass
class Operation(OpenAPIElement):
    responses: Dict[str, Response]
    tags: Optional[List[str]] = None
    operation_id: Optional[str] = None
    summary: Optional[str] = None
    consumes: Optional[List[str]] = None
    produces: Optional[List[str]] = None
    schemes: Optional[List[str]] = None
    description: Optional[str] = None
    external_docs: Optional[ExternalDocs] = None
    parameters: Optional[List[Union[Parameter, Reference]]] = None
    deprecated: Optional[bool] = None
    security: Optional[List[SecurityRequirement]] = None


@dataclass
class PathItem(OpenAPIElement):
    ref: Optional[str] = None
    get: Optional[Operation] = None
    put: Optional[Operation] = None
    post: Optional[Operation] = None
    delete: Optional[Operation] = None
    options: Optional[Operation] = None
    head: Optional[Operation] = None
    patch: Optional[Operation] = None
    parameters: Optional[List[Union[Parameter, Reference]]] = None


class SecurityScheme(OpenAPIElement, ABC):
    """Abstract security scheme"""


@dataclass
class BasicSecurity(SecurityScheme):
    type: SecuritySchemeType = SecuritySchemeType.BASIC
    description: Optional[str] = None


@dataclass
class APIKeySecurity(SecurityScheme):
    name: str
    in_: APIKeyLocation
    type: SecuritySchemeType = SecuritySchemeType.APIKEY
    description: Optional[str] = None


@dataclass
class OAuth2Security(SecurityScheme):
    flow: OAuthFlowType
    scopes: Dict[str, str]
    authorization_url: Optional[str] = None
    token_url: Optional[str] = None
    type: SecuritySchemeType = SecuritySchemeType.OAUTH2
    description: Optional[str] = None


@dataclass
class Tag(OpenAPIElement):
    name: str
    description: Optional[str] = None
    external_docs: Optional[ExternalDocs] = None


@dataclass
class OpenAPI(OpenAPIRoot):
    swagger: str = "2.0"
    info: Optional[Info] = None
    host: Optional[str] = None
    base_path: Optional[str] = None
    schemes: Optional[List[str]] = None
    consumes: Optional[List[str]] = None
    produces: Optional[List[str]] = None
    paths: Optional[Dict[str, PathItem]] = None
    definitions: Optional[Dict[str, Schema]] = None
    parameters: Optional[Dict[str, Parameter]] = None
    responses: Optional[Dict[str, Response]] = None
    security_definitions: Optional[Dict[str, SecurityScheme]] = None
    security: Optional[List[SecurityRequirement]] = None
    tags: Optional[List[Tag]] = None
    external_docs: Optional[ExternalDocs] = None



import asyncio

from nestipy.microservice.client import RedisClientProxy, MicroserviceOption
from nestipy.microservice.client.option import MicroserviceClientOption
from nestipy.microservice.server.base import MicroServiceServer


async def main():
    client = RedisClientProxy(
        MicroserviceOption(
            option=MicroserviceClientOption(
                host='localhost',
                port=6379
            )
        ),
    )
    server = MicroServiceServer(pubsub=client)
    print("Running server ...")
    await server.listen()


if __name__ == "__main__":
    asyncio.run(main())



import enum
from dataclasses import dataclass


class Transport(enum.Enum):
    REDIS = "REDIS"
    MQTT = "MQTT"
    NATS = "NATS"
    RABBITMQ = "RABBITMQ"
    CUSTOM = "CUSTOM"


@dataclass
class MicroserviceClientOption:
    host: str
    port: int



from inspect import getmembers, isfunction
from typing import Union, Type

from nestipy.metadata import ModuleMetadata, Reflect
from nestipy.graphql.meta import NestipyGraphqlKey


class GraphqlExplorer:

    @classmethod
    def explore(cls, module_ref: Union[Type, object]) -> type[list[dict], list[dict], list[dict]]:
        query = []
        mutation = []
        subscription = []
        providers = Reflect.get_metadata(module_ref, ModuleMetadata.Providers, [])
        for provider in providers:
            if not Reflect.get_metadata(provider, NestipyGraphqlKey.resolver, False):
                continue
            cls._explore_resolver(provider, query, mutation, subscription)

        return query, mutation, subscription

    @classmethod
    def _explore_resolver(cls, resolver: Type, query: list,
                          mutation: list,
                          subscription: list):
        members = getmembers(resolver, isfunction)
        for method_name, _ in members:
            if method_name.startswith("__"):
                continue
            method = getattr(resolver, method_name)
            graphql_handler = Reflect.get_metadata(method, NestipyGraphqlKey.handler, None)
            if graphql_handler is not None:
                return_type = Reflect.get_metadata(method, NestipyGraphqlKey.return_type, None)
                data: dict = {
                    'return_type': return_type,
                    'class': resolver,
                    'handler_name': method_name
                }
                match graphql_handler:
                    case NestipyGraphqlKey.query:
                        query.append(data)
                    case NestipyGraphqlKey.mutation:
                        mutation.append(data)
                    case NestipyGraphqlKey.subscription:
                        subscription.append(data)



"""
This module defines classes that can be used to generate OpenAPI Documentation
version 3.
https://swagger.io/specification/
"""
from abc import ABC
from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, List, Optional, Union

from .common import OpenAPIElement
from .common import OpenAPIRoot, normalize_dict


class ParameterLocation(Enum):
    QUERY = "query"
    HEADER = "header"
    PATH = "path"
    COOKIE = "cookie"


class ValueType(Enum):
    ARRAY = "array"
    BOOLEAN = "boolean"
    INTEGER = "integer"
    NUMBER = "number"
    OBJECT = "object"
    STRING = "string"


class ValueFormat(Enum):
    BASE64 = "base64"
    BINARY = "binary"
    BYTE = "byte"
    DATE = "date"
    DATETIME = "date-time"
    DOUBLE = "double"
    FLOAT = "float"
    INT32 = "int32"
    INT64 = "int64"
    PASSWORD = "password"
    EMAIL = "email"
    UUID = "uuid"
    PARTIALTIME = "partial-time"


class SecuritySchemeType(Enum):
    APIKEY = "apiKey"
    HTTP = "http"
    OAUTH = "oauth2"
    OAUTH2 = "oauth2"
    OIDC = "openIdConnect"
    OPENIDCONNECT = "openIdConnect"


@dataclass
class Contact(OpenAPIElement):
    name: Optional[str] = None
    url: Optional[str] = None
    email: Optional[str] = None


@dataclass
class ExternalDocs(OpenAPIElement):
    url: str
    description: Optional[str] = None


@dataclass
class License(OpenAPIElement):
    name: str
    url: Optional[str] = None


@dataclass
class Info(OpenAPIElement):
    title: str
    version: str
    description: Optional[str] = None
    terms_of_service: Optional[str] = None
    contact: Optional[Contact] = None
    license: Optional[License] = None


@dataclass
class ServerVariable(OpenAPIElement):
    default: str
    description: Optional[str] = None
    enum: Optional[List[str]] = None


@dataclass
class Server(OpenAPIElement):
    url: str
    description: Optional[str] = None
    variables: Optional[Dict[str, ServerVariable]] = None


@dataclass
class XML(OpenAPIElement):
    name: Optional[str] = None
    namespace: Optional[str] = None
    prefix: Optional[str] = None
    attribute: Optional[bool] = None
    wrapped: Optional[bool] = None


@dataclass
class Discriminator(OpenAPIElement):
    property_name: str
    mapping: Optional[Dict[str, str]] = None


@dataclass
class Schema(OpenAPIElement):
    type: Union[None, str, ValueType] = None
    format: Union[None, str, ValueFormat] = None
    required: Optional[List[str]] = None
    properties: Optional[Dict[str, Union["Schema", "Reference"]]] = None
    default: Optional[Any] = None
    deprecated: Optional[bool] = None
    example: Any = None
    external_docs: Optional[ExternalDocs] = None
    ref: Optional[str] = None
    title: Optional[str] = None
    max_length: Optional[float] = None
    min_length: Optional[float] = None
    maximum: Optional[float] = None
    minimum: Optional[float] = None
    nullable: Optional[bool] = None
    xml: Optional[XML] = None
    items: Union[None, "Schema", "Reference"] = None
    enum: Optional[List[str]] = None
    discriminator: Optional[Discriminator] = None
    all_of: Optional[List[Union["Schema", "Reference"]]] = None
    any_of: Optional[List[Union["Schema", "Reference"]]] = None
    one_of: Optional[List[Union["Schema", "Reference"]]] = None
    not_: Optional[List[Union["Schema", "Reference"]]] = None


@dataclass
class Header(OpenAPIElement):
    description: Optional[str] = None
    schema: Union[None, Schema, "Reference"] = None


@dataclass
class Example(OpenAPIElement):
    summary: Optional[str] = None
    description: Optional[str] = None
    value: Any = None
    external_value: Optional[str] = None


@dataclass
class Reference(OpenAPIElement):
    ref: str

    def to_obj(self) -> Dict[str, str]:
        return {"$ref": self.ref}


@dataclass
class Encoding(OpenAPIElement):
    content_type: Optional[str] = None
    headers: Optional[Dict[str, Union[Header, Reference]]] = None
    style: Optional[str] = None
    explode: Optional[bool] = None
    allow_reserved: Optional[bool] = None


@dataclass
class Link(OpenAPIElement):
    operation_ref: Optional[str] = None
    operation_id: Optional[str] = None
    parameters: Optional[Dict[str, Any]] = None
    request_body: Any = None
    description: Optional[str] = None
    server: Optional[Server] = None


@dataclass
class MediaType(OpenAPIElement):
    schema: Union[None, Schema, Reference] = None
    example: Any = None
    examples: Optional[Dict[str, Union[Example, Reference]]] = None
    encoding: Optional[Dict[str, Encoding]] = None


@dataclass
class Response(OpenAPIElement):
    description: Optional[str] = None
    headers: Optional[Dict[str, Union[Header, Reference]]] = None
    content: Optional[Dict[str, Union[MediaType, Reference]]] = None
    links: Optional[Dict[str, Union[Link, Reference]]] = None


@dataclass
class Parameter(OpenAPIElement):
    name: str
    in_: ParameterLocation
    schema: Union[None, Schema, Reference] = None
    content: Optional[Dict[str, MediaType]] = None
    description: Optional[str] = None
    required: Optional[bool] = None
    deprecated: Optional[bool] = None
    allow_empty_value: Optional[bool] = None
    example: Optional[Any] = None
    examples: Optional[Dict[str, Union[Example, Reference]]] = None


@dataclass
class RequestBody(OpenAPIElement):
    content: Dict[str, MediaType]
    required: Optional[bool] = None
    description: Optional[str] = None


@dataclass
class SecurityRequirement(OpenAPIElement):
    name: str
    value: List[str]

    def to_obj(self):
        return {self.name: self.value}


@dataclass
class Operation(OpenAPIElement):
    responses: Dict[str, Response]
    tags: Optional[List[str]] = None
    operation_id: Optional[str] = None
    summary: Optional[str] = None
    description: Optional[str] = None
    external_docs: Optional[ExternalDocs] = None
    parameters: Optional[List[Union[Parameter, Reference]]] = None
    request_body: Union[None, RequestBody, Reference] = None
    callbacks: Optional[Dict[str, Union["Callback", Reference]]] = None
    deprecated: Optional[bool] = None
    security: Optional[List[SecurityRequirement]] = None
    servers: Optional[List[Server]] = None


@dataclass
class PathItem(OpenAPIElement):
    summary: Optional[str] = None
    ref: Optional[str] = None
    description: Optional[str] = None
    get: Optional[Operation] = None
    put: Optional[Operation] = None
    post: Optional[Operation] = None
    delete: Optional[Operation] = None
    options: Optional[Operation] = None
    head: Optional[Operation] = None
    patch: Optional[Operation] = None
    trace: Optional[Operation] = None
    servers: Optional[List[Server]] = None
    parameters: Optional[List[Union[Parameter, Reference]]] = None


@dataclass
class Callback(OpenAPIElement):
    expression: str
    path: PathItem

    def to_obj(self):
        return {self.expression: normalize_dict(self.path)}


@dataclass
class OAuthFlow(OpenAPIElement):
    scopes: Dict[str, str]
    authorization_url: Optional[str] = None
    token_url: Optional[str] = None
    refresh_url: Optional[str] = None


@dataclass
class OAuthFlows(OpenAPIElement):
    implicit: Optional[OAuthFlow] = None
    password: Optional[OAuthFlow] = None
    client_credentials: Optional[OAuthFlow] = None
    authorization_code: Optional[OAuthFlow] = None


class SecurityScheme(OpenAPIElement, ABC):
    """Abstract security scheme"""


@dataclass
class HTTPSecurity(SecurityScheme):
    scheme: str
    type: SecuritySchemeType = SecuritySchemeType.HTTP
    description: Optional[str] = None
    bearer_format: Optional[str] = None


@dataclass
class APIKeySecurity(SecurityScheme):
    name: str
    in_: ParameterLocation
    type: SecuritySchemeType = SecuritySchemeType.APIKEY
    description: Optional[str] = None


@dataclass
class OAuth2Security(SecurityScheme):
    flows: OAuthFlows
    type: SecuritySchemeType = SecuritySchemeType.OAUTH2
    description: Optional[str] = None


@dataclass
class OpenIdConnectSecurity(SecurityScheme):
    open_id_connect_url: str
    type: SecuritySchemeType = SecuritySchemeType.OPENIDCONNECT
    description: Optional[str] = None


@dataclass
class Components(OpenAPIElement):
    schemas: Optional[Dict[str, Union[Schema, Reference]]] = None
    responses: Optional[Dict[str, Union[Response, Reference]]] = None
    parameters: Optional[Dict[str, Union[Parameter, Reference]]] = None
    examples: Optional[Dict[str, Union[Example, Reference]]] = None
    request_bodies: Optional[Dict[str, Union[RequestBody, Reference]]] = None
    headers: Optional[Dict[str, Union[Header, Reference]]] = None
    security_schemes: Optional[Dict[str, Union[SecurityScheme, Reference]]] = None
    links: Optional[Dict[str, Union[Link, Reference]]] = None
    callbacks: Optional[Dict[str, Union[Callback, Reference]]] = None


@dataclass
class Tag(OpenAPIElement):
    name: str
    description: Optional[str] = None
    external_docs: Optional[ExternalDocs] = None


@dataclass
class Security(OpenAPIElement):
    requirements: List[SecurityRequirement]
    optional: bool = False

    def to_obj(self):
        items = [normalize_dict(item) for item in self.requirements]
        if self.optional:
            items.insert(0, {})
        return items


@dataclass
class OpenAPI(OpenAPIRoot):
    openapi: str = "3.0.3"
    info: Optional[Info] = None
    paths: Optional[Dict[str, PathItem]] = None
    servers: Optional[List[Server]] = None
    components: Optional[Components] = None
    tags: Optional[List[Tag]] = None
    security: Optional[Security] = None
    external_docs: Optional[ExternalDocs] = None



import enum


class NestipyContextProperty(enum.Enum):
    request: str = '_request'
    response: str = '_response'
    query_params: str = '_query_params'
    params: str = '_params'
    session: str = '_session'
    cookies: str = '_cookies'
    body: str = '_body'
    headers: str = '_headers'
    args: str = '_args'
    files: str = '_files'
    execution_context: str = '_execution_context'
    io_server: str = '_websocket_sever'
    io_client: str = '_socket_sid'
    io_data: str = '_socket_data'



import asyncio
import traceback
from typing import Callable, Type

from rich.style import Style

from nestipy.common.logger import console
from nestipy.core import NestipyApplication, NestipyConfig
from nestipy.microservice.client.base import MicroserviceOption
from nestipy.microservice.client.factory import ClientModuleFactory
from nestipy.microservice.proxy import MicroserviceProxy
from nestipy.microservice.server import MicroServiceServer
from nestipy.microservice.server.module import MicroserviceServerModule


class NestipyMicroservice:
    app: NestipyApplication
    servers: list[MicroServiceServer] = []
    _ms_ready: bool = False
    coroutines: list = []

    def __init__(self, module: Type, option: list[MicroserviceOption]):
        self.root_module = module
        self.option = option
        self.app = NestipyApplication()

    async def ready(self):
        for opt in self.option:
            client_adapter = ClientModuleFactory.create(opt)
            server = MicroServiceServer(
                pubsub=client_adapter,
            )
            self.servers.append(server)
        self.app.init(self.root_module)
        self.app.add_module_root_module(MicroserviceServerModule, _init=True)
        # create all instance if not
        await self.app.setup()
        # get server module
        server_module: MicroserviceServerModule = await self.app.get(MicroserviceServerModule)
        controllers = server_module.discover.get_all_controller()
        for server in self.servers:
            MicroserviceProxy(server).apply_routes(controllers)

        self._ms_ready = True

    async def start(self):
        self.coroutines = []
        # start server listener
        if not self._ms_ready:
            await self.ready()
        for server in self.servers:
            self.coroutines.append(server.listen())
        await asyncio.gather(*self.coroutines)

    async def stop(self):
        # for task in self.coroutines:
        #     task.done()
        for server in self.servers:
            try:
                await server.close()
            except RuntimeError as e:
                traceback.print_exc()
                print(e)
            finally:
                pass

    async def __call__(self, scope: dict, receive: Callable, send: Callable):
        if scope.get('type') == 'lifespan':
            def create_task():
                asyncio.create_task(self.start())

            create_task()
            console.print("[INFO]    Microservice startup completed", style=Style(color="green", bold=True))


class NestipyConnectMicroservice(NestipyMicroservice, NestipyApplication):
    _running: bool = False

    def __init__(self, module: Type, config: NestipyConfig, option: list[MicroserviceOption]):
        NestipyMicroservice.__init__(self, module, option)
        NestipyApplication.__init__(self, config)
        self.app = self

    async def __call__(self, scope: dict, receive: Callable, send: Callable):
        await NestipyApplication.__call__(self, scope, receive, send)

    def start_all_microservices(self):
        def on_start():
            asyncio.create_task(self.start())

        self.app.on_startup(on_start)
        self.app.on_shutdown(self.stop)



import traceback
from typing import Union, Type, Callable

from nestipy.core.guards import GuardProcessor
from nestipy.graphql.graphql_adapter import GraphqlAdapter
from nestipy.ioc import NestipyContainer
from nestipy.ioc import RequestContextContainer
from .graphql_explorer import GraphqlExplorer
from .graphql_module import GraphqlModule, GraphqlOption
from ..common import Request, Response
from ..common.logger import logger
from ..core.adapter.http_adapter import HttpAdapter
from ..core.context.execution_context import ExecutionContext
from ..types_ import NextFn


class GraphqlProxy:

    def __init__(self, adapter: HttpAdapter, graphql_server: GraphqlAdapter):
        self._graphql_server = graphql_server
        self._adapter = adapter
        self.container = NestipyContainer.get_instance()

    def apply_resolvers(self, graphql_module: GraphqlModule, modules: list[Union[Type, object]]):

        for module_ref in modules:
            query, mutation, subscription = GraphqlExplorer.explore(module_ref)
            for q in query:
                name, return_type, resolver = self._create_graphql_query_handler(module_ref, q)
                self._graphql_server.add_query_property(name, return_type, resolver)
            for m in mutation:
                name, return_type, resolver = self._create_graphql_query_handler(module_ref, m)
                self._graphql_server.add_mutation_property(name, return_type, resolver)
            for s in subscription:
                name, return_type, resolver = self._create_graphql_query_handler(module_ref, s)
                self._graphql_server.add_subscription_property(name, return_type, resolver)

        self._create_graphql_request_handler(graphql_module.config or GraphqlOption())

    def _create_graphql_query_handler(self, module_ref: Union[Type, object], meta: dict) -> tuple[str, Type, Callable]:
        resolver: Union[object, Type] = meta['class']
        method_name: str = meta['handler_name']
        method = getattr(resolver, method_name)

        async def graphql_handler(*_args, **kwargs):
            context_container = RequestContextContainer.get_instance()
            execution_context = ExecutionContext(
                self._adapter,
                module_ref,
                resolver,
                getattr(resolver, method_name),
                None,
                None,
                kwargs,
                None

            )
            context_container.set_execution_context(execution_context)
            try:
                # TODO: Refactor with routerProxy
                # create execution context

                #  apply guards
                guard_processor: GuardProcessor = await self.container.get(GuardProcessor)
                can_activate = await guard_processor.process(execution_context)
                if not can_activate[0]:
                    # TODO: is this need to specify return Exception
                    raise Exception(f"Not authorized from guard {can_activate[1]}")

                # perform query request
                result = await self.container.get(resolver, method_name)
                return result
            except Exception as e:
                tb = traceback.format_exc()
                logger.error(e)
                logger.error(tb)
                context_container.destroy()
                return self._graphql_server.raise_exception(e)
            finally:
                context_container.destroy()

        # mutate handle inside adapter
        return_type = self._graphql_server.mutate_handler(method, graphql_handler)
        return method_name, return_type, graphql_handler,

    def _create_graphql_request_handler(self, option: GraphqlOption):
        graphql_path = f"/{option.url.strip('/')}"
        # create graphql handler but using handle of graphql_adapter
        self._adapter.mount(
            graphql_path,
            self._graphql_server.create_graphql_asgi_app(
                option=option,
                schema=self._graphql_server.create_schema()
            ).handle
        )

        async def graphql_redirect(_req: Request, res: Response, _next_fn: NextFn):
            return await res.redirect(f"{graphql_path}/")

        self._adapter.get(graphql_path, graphql_redirect, {})

    @classmethod
    def should_render_graphql_ide(cls, req: Request) -> bool:
        return (
                req.method == "GET"
                and req.query_params.get("query") is None
                and any(supported_header in req.headers.get("accept", "") for supported_header in ("text/html", "*/*"))
        )



class HttpStatusMessages:
    CONTINUE = "Continue"
    SWITCHING_PROTOCOLS = "Switching Protocols"
    PROCESSING = "Processing"
    EARLY_HINTS = "Early Hints"
    OK = "OK"
    CREATED = "Created"
    ACCEPTED = "Accepted"
    NON_AUTHORITATIVE_INFORMATION = "Non-Authoritative Information"
    NO_CONTENT = "No Content"
    RESET_CONTENT = "Reset Content"
    PARTIAL_CONTENT = "Partial Content"
    MULTI_STATUS = "Multi-Status"
    ALREADY_REPORTED = "Already Reported"
    IM_USED = "IM Used"
    MULTIPLE_CHOICES = "Multiple Choices"
    MOVED_PERMANENTLY = "Moved Permanently"
    FOUND = "Found"
    SEE_OTHER = "See Other"
    NOT_MODIFIED = "Not Modified"
    USE_PROXY = "Use Proxy"
    TEMPORARY_REDIRECT = "Temporary Redirect"
    PERMANENT_REDIRECT = "Permanent Redirect"
    BAD_REQUEST = "Bad Request"
    UNAUTHORIZED = "Unauthorized"
    PAYMENT_REQUIRED = "Payment Required"
    FORBIDDEN = "Forbidden"
    NOT_FOUND = "Not Found"
    METHOD_NOT_ALLOWED = "Method Not Allowed"
    NOT_ACCEPTABLE = "Not Acceptable"
    PROXY_AUTHENTICATION_REQUIRED = "Proxy Authentication Required"
    REQUEST_TIMEOUT = "Request Timeout"
    CONFLICT = "Conflict"
    GONE = "Gone"
    LENGTH_REQUIRED = "Length Required"
    PRECONDITION_FAILED = "Precondition Failed"
    PAYLOAD_TOO_LARGE = "Payload Too Large"
    URI_TOO_LONG = "URI Too Long"
    UNSUPPORTED_MEDIA_TYPE = "Unsupported Media Type"
    RANGE_NOT_SATISFIABLE = "Range Not Satisfiable"
    EXPECTATION_FAILED = "Expectation Failed"
    IM_A_TEAPOT = "I'm a teapot"
    MISDIRECTED_REQUEST = "Misdirected Request"
    UNPROCESSABLE_ENTITY = "Unprocessable Entity"
    LOCKED = "Locked"
    FAILED_DEPENDENCY = "Failed Dependency"
    TOO_EARLY = "Too Early"
    UPGRADE_REQUIRED = "Upgrade Required"
    PRECONDITION_REQUIRED = "Precondition Required"
    TOO_MANY_REQUESTS = "Too Many Requests"
    REQUEST_HEADER_FIELDS_TOO_LARGE = "Request Header Fields Too Large"
    UNAVAILABLE_FOR_LEGAL_REASONS = "Unavailable For Legal Reasons"
    INTERNAL_SERVER_ERROR = "Internal Server Error"
    NOT_IMPLEMENTED = "Not Implemented"
    BAD_GATEWAY = "Bad Gateway"
    SERVICE_UNAVAILABLE = "Service Unavailable"
    GATEWAY_TIMEOUT = "Gateway Timeout"
    HTTP_VERSION_NOT_SUPPORTED = "HTTP Version Not Supported"
    VARIANT_ALSO_NEGOTIATES = "Variant Also Negotiates"
    INSUFFICIENT_STORAGE = "Insufficient Storage"
    LOOP_DETECTED = "Loop Detected"
    NOT_EXTENDED = "Not Extended"
    NETWORK_AUTHENTICATION_REQUIRED = "Network Authentication Required"



import logging

from rich.console import Console
from rich.style import Style

logger = logging.getLogger('uvicorn')
console = Console(color_system="truecolor", style=Style(bold=True, color='green'))



from abc import ABC, abstractmethod
from typing import Any, Callable, Type, Union

from .graphql_asgi import GraphqlAsgi
from .graphql_module import GraphqlOption


class GraphqlAdapter(ABC):
    _query_properties: list = []
    _mutation_properties: list = []
    _subscription_properties: list = []

    @abstractmethod
    def create_query_field_resolver(self, resolver: Callable) -> object:
        pass

    @abstractmethod
    def raise_exception(self, e: Exception):
        pass

    @abstractmethod
    def create_mutation_field_resolver(self, resolver: Callable) -> object:
        pass

    @abstractmethod
    def create_subscription_field_resolver(self, resolver: Callable) -> object:
        pass

    def add_query_property(self, property_name: str, return_type: Any, resolver: Callable):
        self._query_properties.append((property_name, self.create_query_field_resolver(resolver)))

    def add_mutation_property(self, property_name: str, return_type: Any, resolver: Callable):
        self._mutation_properties.append((property_name, self.create_mutation_field_resolver(resolver)))

    def add_subscription_property(self, property_name: str, return_type: Any, resolver: Callable):
        self._subscription_properties.append((property_name, self.create_subscription_field_resolver(resolver)))

    def create_query(self, name: str = 'Query') -> Union[Type, None]:
        if len(self._query_properties) == 0:
            return None
        return type(name, (), {k: r for (k, r) in self._query_properties})

    def create_mutation(self, name: str = 'Mutation') -> Union[Type, None]:
        if len(self._mutation_properties) == 0:
            return None
        return type(name, (), {k: r for (k, r) in self._mutation_properties})

    def create_subscription(self, name: str = 'Subscription') -> Union[Type, None]:
        if len(self._subscription_properties) == 0:
            return None
        return type(name, (), {k: r for (k, r) in self._subscription_properties})

    @abstractmethod
    def create_schema(self) -> Any:
        pass

    @abstractmethod
    def mutate_handler(self, original_handler: Callable, wrapper_handler: Callable) -> Type:
        pass

    @abstractmethod
    def create_graphql_asgi_app(self, schema: Any, option: GraphqlOption) -> GraphqlAsgi:
        pass



from typing import Any, Union

from .reflect import Reflect


class SetMetadata:
    def __init__(self, key: str, data: Any, as_list: bool = False, as_dict: bool = False):
        self.key = key
        self.data = data
        self.as_list = as_list
        self.as_dict = as_dict
        pass

    def __call__(self, cls):
        default = [] if self.as_list else {} if self.as_dict else None
        meta: Union[list, dict] = Reflect.get_metadata(cls, self.key, default)
        if self.as_list:
            meta = meta + self.data
        elif self.as_dict:
            meta.update(self.data)
        else:
            meta = self.data
        Reflect.set_metadata(cls, self.key, meta)
        return cls



import inspect
import traceback
from typing import Type, Callable, cast, Any

from nestipy.common.exception.http import HttpException
from nestipy.common.logger import logger
from nestipy.core.context.execution_context import ExecutionContext
from nestipy.core.exception.processor import ExceptionFilterHandler
from nestipy.core.guards.processor import GuardProcessor
from nestipy.core.interceptor.processor import RequestInterceptor
from nestipy.ioc import RequestContextContainer, NestipyContainer
from nestipy.metadata.class_ import ClassMetadata
from nestipy.metadata.reflect import Reflect
from .context import RpcRequest, RpcException
from .decorator import MICROSERVICE_LISTENER, MicroserviceMetadata, MicroserviceData
from .exception import RPCErrorCode, RPCErrorMessage
from .server.base import MicroServiceServer


class MicroserviceProxy:

    def __init__(self, adapter: MicroServiceServer):
        self.adapter = adapter

    def apply_routes(self, controllers: list[object | type]):
        for ctrl in controllers:
            elements = inspect.getmembers(ctrl, lambda a: inspect.isfunction(a) or inspect.iscoroutinefunction(a))
            methods = [
                method for (method, _) in elements
                if not method.startswith("__") and self._is_listener(getattr(ctrl, method))
            ]
            for m in methods:
                metadata: ClassMetadata = Reflect.get_metadata(ctrl, ClassMetadata.Metadata, None)
                if metadata is not None:
                    data: MicroserviceData = Reflect.get_metadata(
                        getattr(ctrl, m), MICROSERVICE_LISTENER,
                        None
                    )
                    if not data.transport or data.transport == self.adapter.get_transport():
                        handler = self._create_event_handler(
                            metadata.get_module(),
                            ctrl,
                            m,
                        )
                        self.register_listener(handler, data)

    @classmethod
    def _is_listener(cls, method: Callable):
        return Reflect.get_metadata(method, MICROSERVICE_LISTENER, None) is not None

    def register_listener(self, listener: Callable, data: MicroserviceData):
        match data.type:
            case MicroserviceMetadata.Event:
                self.adapter.subscribe(data.pattern, listener)
            case MicroserviceMetadata.Message:
                self.adapter.request_subscribe(data.pattern, listener)

    @classmethod
    def _create_event_handler(
            cls, module_ref: Type,
            controller: Type,
            method_name: str,
    ):
        async def event_handler(server: MicroServiceServer, request: RpcRequest):
            context_container = RequestContextContainer.get_instance()
            container = NestipyContainer.get_instance()
            execution_context = ExecutionContext(
                None,
                module_ref,
                controller,
                getattr(controller, method_name),
                cast(Any, request),
                None,
                None,
                None,
                server,
                None,
                request.data
            )
            context_container.set_execution_context(execution_context)
            try:
                guard_processor: GuardProcessor = await container.get(GuardProcessor)
                can_activate = await guard_processor.process(execution_context, False)
                if not can_activate[0]:
                    raise RpcException(RPCErrorCode.PERMISSION_DENIED, RPCErrorMessage.PERMISSION_DENIED)

                async def next_fn():
                    return await container.get(controller.__class__, method_name)

                interceptor: RequestInterceptor = await container.get(RequestInterceptor)
                return await interceptor.intercept(execution_context, next_fn, False)

            except Exception as e:
                tb = traceback.format_exc()
                logger.error(e)
                logger.error(tb)
                if not isinstance(e, RpcException):
                    e = RpcException(RPCErrorCode.INTERNAL, RPCErrorMessage.INTERNAL)
                exception_handler: ExceptionFilterHandler = await container.get(ExceptionFilterHandler)
                return await exception_handler.catch(e, execution_context, False)

        return event_handler



from abc import ABC, abstractmethod
from typing import Type, Union, Callable, TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from nestipy.core.adapter.http_adapter import HttpAdapter
    from nestipy.common.http_ import Request, Response


class ArgumentHost(ABC):

    def __init__(
            self,
            adapter: Union["HttpAdapter", None],
            module: Union[Type, object],
            class_handler: Union[Type, object],
            handler: Callable,
            req: Union["Request", None],
            res: Union["Response", None],
            graphql_args: Optional[dict] = None,
            graphql_context: Optional[any] = None,
            socket_server:  Optional[any] = None,
            socket_client:  Optional[any] = None,
            socket_data:  Optional[any] = None,
    ):
        self._adapter = adapter
        self._module = module
        self._handler = handler
        self._class_handler = class_handler
        self._req = req
        self._res = res
        self._graphql_args = graphql_args or {}
        self._graphql_context = graphql_context or None
        self._socket_server = socket_server
        self._socket_client = socket_client
        self._socket_data = socket_data

    def get_adapter(self) -> Union["HttpAdapter", None]:
        return self._adapter

    def get_module(self) -> Union[Type, object, None]:
        return self._module

    def get_request(self) -> Union["Request", None]:
        return self._req

    def get_response(self) -> "Response":
        return self._res

    def get_class(self) -> Union[Type, None]:
        return self._class_handler

    def get_handler(self) -> Union[Callable, None]:
        return self._handler

    @abstractmethod
    def get_args(self):
        pass



from typing import Type, Callable, Union

from nestipy.common.http_ import Response, Request
from .argument_host import ArgumentHost
from .graphql_argument_host import GraphqlArgumentHost
from .http_argument_host import HttpArgumentHost
from .rpc_argumnt_host import RpcArgumentHost
from .websocket_argument_host import WebsocketArgumentHost


class ExecutionContext(ArgumentHost):
    def get_args(self) -> tuple[Union[Type, object], Callable, Union[Request, None], Union[Response, None]]:
        return self.get_class(), self.get_handler(), self.get_request(), self.get_response()

    def switch_to_http(self) -> HttpArgumentHost:
        return HttpArgumentHost(
            self.get_adapter(),
            self.get_module(),
            self.get_class(),
            self.get_handler(),
            self.get_request(),
            self.get_response()
        )

    def switch_to_graphql(self) -> GraphqlArgumentHost:
        return GraphqlArgumentHost(
            self.get_adapter(),
            self.get_module(),
            self.get_class(),
            self.get_handler(),
            self.get_request(),
            self.get_response(),
            self._graphql_args,
            self._graphql_context,
        )

    def switch_to_websocket(self) -> WebsocketArgumentHost:
        return WebsocketArgumentHost(
            self.get_adapter(),
            self.get_module(),
            self.get_class(),
            self.get_handler(),
            self.get_request(),
            self.get_response(),
            None,
            None,
            self._socket_server,
            self._socket_client,
            self._socket_data,
        )

    def switch_to_rpc(self) -> RpcArgumentHost:
        return RpcArgumentHost(
            None,
            self.get_module(),
            self.get_class(),
            self.get_handler(),
            self.get_request(),
            None,
            None,
            None,
            self._socket_data
        )

    def get_context(self):
        return self



from .adapter.http_adapter import HttpAdapter
from .constant import AppKey
from .context.execution_context import ExecutionContext, ArgumentHost, HttpArgumentHost
from .discover import DiscoverService
from .middleware import MiddlewareConsumer
from .nestipy_application import NestipyApplication, NestipyConfig
from .nestipy_factory import NestipyFactory
from .on_init import OnInit
from .on_destroy import OnDestroy
from .platform import NestipyFastApiApplication, NestipyBlackSheepApplication
from .template import MinimalJinjaTemplateEngine
from .nestipy_microservice import NestipyMicroservice, NestipyConnectMicroservice

__all__ = [
    "NestipyFactory",
    "NestipyApplication",
    "NestipyConfig",
    "HttpAdapter",
    "ExecutionContext",
    "ArgumentHost",
    "HttpArgumentHost",
    "NestipyFastApiApplication",
    "NestipyBlackSheepApplication",
    "MinimalJinjaTemplateEngine",
    "AppKey",
    "MiddlewareConsumer",
    "DiscoverService",
    "OnInit",
    "OnDestroy",
    "NestipyMicroservice",
    "NestipyConnectMicroservice"
]



import asyncio

from nestipy.microservice.client import MqttClientProxy, MicroserviceOption
from nestipy.microservice.client.option import MicroserviceClientOption


async def main():
    client = MqttClientProxy(
        MicroserviceOption(
            option=MicroserviceClientOption(
                host='localhost',
                port=1883
            )
        ),
    )
    await client.send("topic", "request")


if __name__ == "__main__":
    asyncio.run(main())



import uuid

APP_GUARD = '__APP__GUARD__'
APP_FILTER = '__APP__FILTER__'
APP_INTERCEPTOR = '__APP__INTERCEPTOR__'


class _AppKey:

    @property
    def APP_GUARD(self):
        return APP_GUARD + uuid.uuid4().hex

    @property
    def APP_FILTER(self):
        return APP_FILTER + uuid.uuid4().hex

    @property
    def APP_INTERCEPTOR(self):
        return APP_INTERCEPTOR + uuid.uuid4().hex


AppKey = _AppKey()



from typing import Callable, Type, Union, Any

from .class_ import Module, Controller, Injectable
from .method import Route, Get, Post, Patch, Put, Delete


def apply_decorators(*callback: Callable[[Union[Type | Callable]], Any]):
    def decorator(cls: Union[Type | Callable]):
        for deco in callback:
            cls = deco(cls)
        return cls

    return decorator


__all__ = [
    "Module",
    "Controller",
    "Injectable",
    "Route",
    "Get",
    "Post",
    "Put",
    "Patch",
    "Delete",
    "apply_decorators"
]



from nestipy.ioc.dependency import Req, SocketData, Inject

Payload = SocketData
Ctx = Req


def Client(token: str = None):
    return Inject(token)


__all__ = [
    "Payload",
    "Ctx",
    "Client"
]



from .event_module import EventEmitterModule
from .event_decorator import OnEvent, OnceEvent
from .event_emitter import EventEmitter

__all__ = [
    "EventEmitterModule",
    "OnEvent",
    "OnceEvent",
    "EventEmitter"
]



from typing import Type, Callable, Union

from nestipy.common import Module
from nestipy.core.nestipy_factory import NestipyFactory
from nestipy.core.platform import NestipyBlackSheepApplication, NestipyFastApiApplication
from nestipy.dynamic_module import DynamicModule
from nestipy.ioc import NestipyContainer
from nestipy.ioc.provider import ModuleProviderDict
from .client import TestClient


class TestingModuleRef:
    __test__ = False

    def __init__(self, root_module: Type):
        self._root_module = root_module

    async def get(self, key: any):
        await NestipyFactory.create(self._root_module).ready()
        return await NestipyContainer.get_instance().get(key)

    def create_nestipy_client(
            self,
            platform: Union[Type[NestipyBlackSheepApplication], Type[
                NestipyFastApiApplication]] = NestipyFastApiApplication
    ) -> TestClient:
        return TestClient(NestipyFactory[platform].create(self._root_module))


class Test:
    __test__ = False

    @staticmethod
    def create_testing_module(
            providers: list[Callable | ModuleProviderDict] = None,
            controllers: list[Callable] = None,
            imports: list[Union[Type, Callable, ModuleProviderDict, DynamicModule]] = None,
            exports: list[Union[Type, Callable, str]] = None,
            is_global: bool = False
    ) -> TestingModuleRef:
        @Module(
            providers=providers,
            controllers=controllers,
            imports=imports,
            exports=exports,
            is_global=is_global
        )
        class _RootTestingModule:
            pass

        return TestingModuleRef(_RootTestingModule)



import dataclasses
import inspect
from enum import Enum
from typing import Union, Optional, Any, Type, Callable

from pydantic import TypeAdapter, BaseModel

from nestipy.metadata import SetMetadata
from .openapi_docs.v3 import RequestBody, Response, Parameter
from .openapi_docs.v3 import SecurityRequirement, MediaType, Schema


class ApiConsumer(Enum):
    JSON = "application/json"
    MULTIPART = "multipart/form-data"
    X_WWW_FORM_URLENCODED = "application/x-www-form-urlencoded"


def _dict_to_dataclass(data: dict[str, Any], cls: Any) -> Any:
    field_names = {field.name for field in dataclasses.fields(cls)}
    filtered_data = {key: value for key, value in data.items() if key in field_names}
    return cls(**filtered_data)


def _get_json_of_body(body: Union[BaseModel, Type]):
    content: dict = {}
    ref_template = "#/components/schemas/{model}"
    if dataclasses.is_dataclass(body):
        content: dict = TypeAdapter(body).json_schema(ref_template=ref_template)
    elif inspect.isclass(body) and issubclass(body, BaseModel):
        content: dict = body.model_json_schema(ref_template=ref_template)
    return content


def _create_decorator_to_add_schema_ref(decorator: callable, content: dict):
    if "$defs" not in content:
        return decorator

    def update_decorator(cls: Union[Type, Callable]):
        new_cls = decorator(cls)
        class_decorated = SetMetadata(key='__openapi__schemas', as_dict=True, data=content["$defs"])(new_cls)
        return class_decorated

    return update_decorator


def ApiBody(body: Union[BaseModel, Type] = None, consumer: ApiConsumer = ApiConsumer.JSON):
    content = _get_json_of_body(body)
    body = RequestBody(
        required=True,
        content={
            consumer.value: MediaType(
                schema=_dict_to_dataclass(content, Schema),
            )
        },
    )
    decorator = SetMetadata(key='__openapi__request_body', data=body)
    return _create_decorator_to_add_schema_ref(decorator, content)


def ApiResponse(
        status: int,
        response: Union[BaseModel, Type] = None,
        description: str = None,
        example: Any = None,
        consumer: ApiConsumer = ApiConsumer.JSON
):
    content = _get_json_of_body(response)
    response_body = Response(
        description=description,
        content={
            consumer.value: MediaType(
                schema=_dict_to_dataclass(content, Schema),
                example=example
            )
        },
    )
    return _ApiResponse(status, response_body)


def _ApiResponse(status: int, response: Response):
    return SetMetadata(key='__openapi__responses', data={status: response}, as_dict=True)


def NoSwagger():
    return SetMetadata(key='__openapi__no_swagger', data=True)


def ApiOkResponse(schema: Union[BaseModel, Type] = None, description: Optional[str] = None, example: Any = None):
    return ApiResponse(
        status=200,
        description=description,
        response=schema,
        example=example
    )


def ApiCreatedResponse(schema: Union[BaseModel, Type] = None, description: Optional[str] = None, example: Any = None):
    return ApiResponse(
        status=201,
        description=description,
        response=schema,
        example=example
    )


def ApiNotFoundResponse(schema: Union[BaseModel, Type] = None, description: Optional[str] = None, example: Any = None):
    return ApiResponse(
        status=404,
        description=description,
        response=schema,
        example=example
    )


def ApiParameter(param: Parameter):
    return SetMetadata(key='__openapi__parameters', data=[param], as_list=True)


def ApiTags(tags: Union[str, list[str]]):
    if isinstance(tags, str):
        tags = [tags]
    return SetMetadata(key='__openapi__tags', data=tags, as_list=True)


def ApiId(api_id: str):
    return SetMetadata(key='__openapi__operation_id', data=api_id)


def ApiBasicAuth():
    return SetMetadata(key='__openapi__security', data=[SecurityRequirement(
        name='basicAuth',
        value=[
            'basicAuth'
        ]
    )], as_list=True)


def ApiBearerAuth():
    return SetMetadata(key='__openapi__security', data=[SecurityRequirement(
        name='bearer',
        value=[
            'bearer'
        ]
    )], as_list=True)


def ApiSecurity(security: SecurityRequirement):
    return SetMetadata(key='__openapi__security', data=[security], as_list=True)

# TODO : Add more openapi decorator



from .base import ClientProxy, MicroserviceOption
from .mqtt import MqttClientProxy
from .redis import RedisClientProxy
from .rabbitmq import RabbitMQClientProxy
from .nats import NatsClientProxy
from .module import ClientsModule, ClientsConfig
from .factory import ClientModuleFactory

__all__ = [
    "ClientProxy",
    "MicroserviceOption",
    "RedisClientProxy",
    "MqttClientProxy",
    "RabbitMQClientProxy",
    "NatsClientProxy",
    "ClientsModule",
    "ClientsConfig",
    "ClientModuleFactory"
]



import base64
import copy
from abc import ABC, abstractmethod
from dataclasses import asdict, fields, is_dataclass
from datetime import date, datetime, time
from enum import Enum
from typing import Any, List, Tuple
from uuid import UUID

import yaml
from essentials.json import dumps

from .mk.contents import OADJSONEncoder


class Format(Enum):
    YAML = "YAML"
    JSON = "JSON"


class OpenAPIElement:
    """Base class for all OpenAPI Elements"""


class OpenAPIRoot(OpenAPIElement):
    """Base class for a root OpenAPI Documentation"""


class ValueTypeHandler(ABC):
    @abstractmethod
    def normalize(self, value: Any) -> Any:
        """Normalizes a value of the given type into another type."""


class CommonBuiltInTypesHandler(ValueTypeHandler):
    def normalize(self, value: Any) -> Any:
        if isinstance(value, UUID):
            return str(value)

        if isinstance(value, Enum):
            return value.value

        if isinstance(value, time):
            return value.strftime("%H:%M:%S")

        if isinstance(value, datetime):
            return value.isoformat()

        if isinstance(value, date):
            return value.strftime("%Y-%m-%d")

        if isinstance(value, bytes):
            return base64.urlsafe_b64encode(value).decode("utf8")

        return value


TYPES_HANDLERS = [CommonBuiltInTypesHandler()]


def normalize_key(key: Any) -> str:
    if isinstance(key, Enum):
        return key.value
    if not isinstance(key, str):
        return key
    first, *others = key.rstrip("_").split("_")
    return "".join([first.lower(), *map(str.title, others)])


def normalize_dict_factory(items: List[Tuple[Any, Any]]) -> Any:
    data = {}
    for key, value in items:
        if value is None:
            continue

        if hasattr(value, "to_obj"):
            value = value.to_obj()

        if key == "ref":
            data["$ref"] = value
            continue

        for handler in TYPES_HANDLERS:
            value = handler.normalize(value)

        data[normalize_key(key)] = value
    return data


def regular_dict_factory(items: List[Tuple[Any, Any]]) -> Any:
    data = {}
    for key, value in items:
        for handler in TYPES_HANDLERS:
            value = handler.normalize(value)

        data[key] = value
    return data


# replicates the asdict method from dataclasses module, to support
# bypassing "asdict" on child properties when they implement a `to_obj`
# method: some entities require a specific shape when represented
def _asdict_inner(obj, dict_factory):
    if hasattr(obj, "to_obj"):
        return obj.to_obj()
    if isinstance(obj, OpenAPIElement):
        result = []
        for f in fields(obj):
            value = _asdict_inner(getattr(obj, f.name), dict_factory)
            result.append((f.name, value))
        return dict_factory(result)
    if hasattr(obj, "model_dump") and callable(obj.model_dump):
        # For Pydantic 2
        return obj.model_dump()
    if hasattr(obj, "dict") and callable(obj.dict):
        # For Pydantic 1
        return obj.dict()
    if is_dataclass(obj):
        return asdict(obj, dict_factory=regular_dict_factory)
    elif isinstance(obj, (list, tuple)):
        return type(obj)(_asdict_inner(v, dict_factory) for v in obj)
    elif isinstance(obj, dict):
        return type(obj)(
            (_asdict_inner(k, dict_factory), _asdict_inner(v, dict_factory))
            for k, v in obj.items()
        )
    else:
        return copy.deepcopy(obj)


def normalize_dict(obj):
    if hasattr(obj, "dict") and callable(obj.dict):
        return obj.dict()
    if hasattr(obj, "to_obj"):
        return obj.to_obj()
    if isinstance(obj, OpenAPIElement):
        return _asdict_inner(obj, dict_factory=normalize_dict_factory)
    return asdict(obj, dict_factory=regular_dict_factory)


class Serializer:
    """
    Provides methods to serialize dataclasses to JSON and YAML.
    """

    def _get_item_dictionary(self, item: Any) -> Any:
        if hasattr(item, "to_obj"):
            assert callable(item.to_obj), "to_obj must be a method"
            return item.to_obj()
        if is_dataclass(item):
            return normalize_dict(item)
        raise TypeError("Expected a dataclass or a class having a `to_obj()` method.")

    def to_obj(self, item: Any) -> Any:
        return self._get_item_dictionary(item)

    def to_json(self, item: Any) -> str:
        return dumps(self.to_obj(item), indent=4, cls=OADJSONEncoder)

    def to_yaml(self, item: Any) -> str:
        rep = yaml.dump(
            self.to_obj(item), sort_keys=False, indent=4, allow_unicode=True
        )
        assert isinstance(rep, str)
        return rep



from typing import Type

from nestipy.ioc import ModuleProviderDict
from nestipy.metadata import Reflect, ClassMetadata, ModuleMetadata


class SpecialProviderExtractor:

    @classmethod
    def extract_special_providers(
            cls,
            module_class: Type,
            subclass: Type,
            key: str
    ) -> list[Type]:
        providers = []
        class_providers = Reflect.get_metadata(module_class, ModuleMetadata.Providers, [])
        metadata: ClassMetadata = Reflect.get_metadata(module_class, ClassMetadata.Metadata, None)
        if metadata is not None:
            class_providers += metadata.get_global_providers()
        for p in class_providers:
            if (isinstance(p, ModuleProviderDict) and
                    isinstance(p.token, str) and p.token.startswith(key) and p.use_class is not None):
                if issubclass(p.use_class, subclass):
                    providers.append(p.use_class)
        return providers



from typing import Union

from openapidocs.mk.v3 import OpenAPIV3DocumentationHandler
from openapidocs.utils.source import read_from_source


def generate_document(source: str, destination: str, style: Union[int, str]):
    # Note: if support for more kinds of OAD versions will be added, handle a version
    # parameter in this function

    data = read_from_source(source)
    handler = OpenAPIV3DocumentationHandler(data, style=style, source=source)

    html = handler.write()

    # TODO: support more kinds of destinations
    with open(destination, encoding="utf8", mode="wt") as output_file:
        output_file.write(html)
